__all__ = ["MoRobtrol", "mo_ant", "mo_half_cheetah", "mo_hopper", "mo_hopper_m3", "mo_humanoid",
           "mo_humanoidstandup", "mo_inverted_double_pendulum", "mo_pusher", "mo_reacher",
           "mo_swimmer", "mo_walker2d"]

from evomo.problems.neuroevolution.morobtrol_pro import MoRobtrol

from .morobtrol import (
    mo_ant,
    mo_half_cheetah,
    mo_hopper,
    mo_hopper_m3,
    mo_humanoid,
    mo_humanoidstandup,
    mo_inverted_double_pendulum,
    mo_pusher,
    mo_reacher,
    mo_swimmer,
    mo_walker2d,
)
