__all__ = [
    "ZDT1",
    "ZDT2",
    "ZDT3",
    "ZDT4",
    "ZDT6",
    "MAF1",
    "MAF2",
    "MAF3",
    "MAF4",
    "MAF5",
    "MAF6",
    "MAF7",
    "MAF8",
    "MAF9",
    "MAF10",
    "MAF11",
    "MAF12",
    "MAF13",
    "MAF14",
    "MAF15",
    "LSMOP1",
    "LSMOP2",
    "LSMOP3",
    "LSMOP4",
    "LSMOP5",
    "LSMOP6",
    "LSMOP7",
    "LSMOP8",
    "LSMOP9",
    "DTLZ1",
    "DTLZ2",
    "DTLZ3",
    "DTLZ4",
    "DTLZ5",
    "DTLZ6",
    "DTLZ7",
]

from evomo.problems.numerical.dtlz import DTLZ1, DTLZ2, DTLZ3, DTLZ4, DTLZ5, DTLZ6, DTLZ7
from evomo.problems.numerical.lsmop import LSMOP1, LSMOP2, LSMOP3, LSMOP4, LSMOP5, LSMOP6, LSMOP7, LSMOP8, LSMOP9
from evomo.problems.numerical.maf import (
    MAF1,
    MAF2,
    MAF3,
    MAF4,
    MAF5,
    MAF6,
    MAF7,
    MAF8,
    MAF9,
    MAF10,
    MAF11,
    MAF12,
    MAF13,
    MAF14,
    MAF15,
)
from evomo.problems.numerical.zdt import ZDT1, ZDT2, ZDT3, ZDT4, ZDT6
